// Copyright 2001 by Columbia University. All rights reserved.

// This class allows one to get information about and control a NAL audio device.
// An audio device can consist of several lower level abstractions depending on
// the OS's implementation. The intent is to create uniform interface so that 
// the user does not care if, for example, the volume is implemented through  
// a mixer device or just a setting on the audio/wave device itself.

#ifndef AUDIODEVICE_H
#define AUDIODEVICE_H

#include <AudioDevices/DeviceCaps.h>

class AudioDevice
{
  public:

  // 16 bit signed
  typedef short sample_t;
 
  // volume range 0-100, so max is 100
  static const int MAX_VOLUME;

  // The default block size will be used to set up buffer size, if the device
  // can not figure out how to do this on its own. Read/Write buffer sizes can
  // be changed when the device is opened.
  static const int DEFAULT_BLOCK_SIZE;

  // The caller of this function is expected to use the DeviceCaps class to
  // find the capabilities of the low level device and then choose how to
  // initialize that device.
  static DeviceCaps* getDevCaps();

  // A DevCap class should describe how the user wants the device to be
  // intialized. The user should call getDevCaps if they want to know what
  // the device supports.
  AudioDevice();
  ~AudioDevice();

  // The funciton getDevCaps() on the low level driver should be called
  // to figure out what the device can support before trying to open the
  // device. If the device does not support the requested audioFormat, the
  // device will not open correctly. The open function will throw an
  // AudioDeviceException, if the device could not be open successfully.
  virtual void open(DeviceCaps *pDeviceCaps)=0;

  // This funtion returns true if the device is open and false if the
  // device is not open.
  virtual bool isOpen()=0;

  // Closes the device. The close function will throw an AudioDeviceException,
  // if the device could not be closed successfully.
  virtual void close()=0;

  // The read function reads data from the audio device's lower level
  // device. The buffer does not have to be of the same size as the
  // block size. The first argument is the buffer to be filled and
  // the second argument is the size of the buffer. The type of data
  // the buffer is filled with depends on the information the device
  // was open with using DeviceCaps.
  // buffSize is in bytes and the number of bytes read is returned.
  virtual int read(AudioDevice::sample_t *appBuff, int buffSize)=0;

  // The write function writes data to the audio device's lower level
  // device. The buffer does not have to be of the same size as the
  // block size. The first argument is the buffer to be filled and
  // the second argument is the size of the buffer. The type of data
  // the buffer is filled with depends on the information the device
  // was open with using DeviceCaps. 
  // buffSize is in bytes and the number of bytes read is returned.
  virtual int write(AudioDevice::sample_t *appBuff, int buffSize)=0;

  // Sets the volume on the input (mirophone) device. Volume can be 0-100.
  virtual void setInputVol(int vol)=0;

  // Returns the input (mirophone) device volume current setting. Volume
  // can be 0-100.
  virtual int getInputVol()=0;

  // Sets the volume on the output (speakers) device. Volume can be 0-100. 
  virtual void setOutputVol(int vol)=0;

  // Returns the output (speakers) device volume current setting. Volume
  // can be 0-100.
  virtual int getOutputVol()=0;

  // Returns the present sample rate. Can be called any time after the
  // device has been opened and is still open.
  virtual AudioFormat::SampleRate getSampleRate()=0;

  // Returns the present number of channels. Can be called any time after
  // the device has been opened and is still open.
  virtual AudioFormat::Channel getNumChannels()=0;

  // Returns the present size of the read blocks being used by the device.
  // Can be called any time after the device has been opened and is still open.
  virtual int getReadBlockSize()=0;

  // Returns the present size of the read blocks being used by the device.
  // Can be called any time after the device has been opened and is still open.
  virtual int getWriteBlockSize()=0;

  protected:

  // true if the device is open
  // false if the device is closed
  bool deviceOpen;

  // The device chosen by the library's user as the input device.
  DevHandle_t inputDeviceToUse;

  // The device chosen by the library's user as the output device.
  DevHandle_t outputDeviceToUse;

  // The size of the blocks used by the read function.
  int readBlockSize;

  // The size of the blocks used by the write function.
  int writeBlockSize;

  private:

};

#endif