// Copyright 2001 by Columbia University. All rights reserved.

// The device capabilities class.

// This class is used as a device information storage class. It is handed to 
// the low level device and that low level device fills out information in
// this class and hands it back to the caller. The caller then chooses what
// capabilities it wants to initialize the device from the capabilities the
// device communicates are supported (through the DevCaps class). The basic
// unit of storage for this class is a DevCap. Some functions in this class
// are meant to be used by the device and other are meant to be used by the
// library's user.

#ifndef DEVICECAPS_H
#define DEVICECAPS_H

#include <AudioDevices/DevCap.h>

using namespace std;

class DeviceCaps
{
  public:
  
  DeviceCaps();
  DeviceCaps(const DeviceCaps &right);
  ~DeviceCaps();

  // Returns the number of devices in the DeviceCaps class. After the
  // AudioDevice::getDevCaps() function has been called the number of
  // devices the platform supports will be returned.
  int getSize();

  // These functions are intended to be called by the library's user to 
  // determine what the device supports or tell the device how to initialize
  // itself.

  // Returns the name of the device according to the low level device.
  // The library's user can call this function to check the name of the
  // device.
  string* getDevName(DevHandle_t handle);

  // The library's user can call this function to get the AudioFormats
  // class and check to see what formats the device supports. Check
  // AudioFormats.h for information on how to interogate the AudioFormats
  // class.
  AudioFormats* getAudioFormats(DevHandle_t handle);

  // Returns true, if the device supports input volume.
  // Returns false, if the device does not support input volume.
  // The library's user can call this function to check if the device
  // supports changing the input volume.
  bool getInVolSupp(DevHandle_t handle);

  // Returns true, if the device supports output volume.
  // Returns false, if the device does not support output volume.
  // The library's user can call this function to check if the device
  // supports changing the output volume.  
  bool getOutVolSupp(DevHandle_t handle);
 
  // Setting this tells the audio device which device to use as the
  // input device. There can only be one input device set per "run"
  // of the deivce. Each time the device is opened the input device
  // can be changed.
  void setInputDeviceToUse(DevHandle_t handle);

  // Setting this tells the audio device which device to use as the
  // output device. There can only be one output device set per "run"
  // of the deivce. Each time the device is opened the input device
  // can be changed.
  void setOutputDeviceToUse(DevHandle_t handle);

  // These "ToUse" functions are intended to be called by the library's user
  // in order to tell the device how it should initialize itself. The 
  // assumtion is that there is only one set of settings to use per "run"
  // of the device. Every time the device is opened the settings can be 
  // changed. One should refer to AudioFormat to better understand what is
  // being set by the following functions.
  void setSampleRateToUse(DevHandle_t handle,
                          AudioFormat::SampleRate sampleRate);
  void setEncodingToUse(DevHandle_t handle,
                        AudioFormat::Encoding encoding);
  void setChannelToUse(DevHandle_t handle,
                       AudioFormat::Channel channel);
  void setNumBitsToUse(DevHandle_t handle,
                       AudioFormat::NumBits numBits);

  // These functions are intended to be used by the device to fill out
  // information about the device in the DeviceCaps class.

  // Sets the name of the device. The device's name is set by the
  // platform's device according to the platform's device layer.
  void setDevName(DevHandle_t handle, const char* name);

  // Sets the AudioFormats class of this DevCaps. The platform's device
  // is interogated by the NAL's platform independent device layer. Setting
  // the AudioFormats is done by the device layer.
  void setAudioFormats(DevHandle_t handle, AudioFormats *audioFormats);

  // Set to true, if the device supports input volume.
  // Set to false, if the device does not support input volume.
  // Setting if the input volume changing is supported is done by the
  // device layer.
  void setInVolSupp(DevHandle_t handle, bool supported);

  // Set to true, if the device supports output volume.
  // Set to false, if the device does not support output volume.
  // Setting if the output volume changing is supported is done by the
  // device layer.
  void setOutVolSupp(DevHandle_t handle, bool supported);

  // The device uses these "ToUse" functions to decide how to initialize
  // itself. One should refer to AudioFormat to better understand what is
  // being set by the following functions. 
  AudioFormat::SampleRate getSampleRateToUse(DevHandle_t handle);
  AudioFormat::Encoding getEncodingToUse(DevHandle_t handle);
  AudioFormat::Channel getChannelToUse(DevHandle_t handle);
  AudioFormat::NumBits getNumBitsToUse(DevHandle_t handle);

  // These functions could be called by the user and are called by the device.

  // Returns the input device handle of the device set for use the next time
  // the device is opened. If the device is open, then the handle returned
  // will be the handle of the input device currently being used.
  DevHandle_t getInputDeviceToUse();

  // Returns the output device handle of the device set for use the next time
  // the device is opened. If the device is open, then the handle returned
  // will be the handle of the output device currently being used.
  DevHandle_t getOutputDeviceToUse();

  // The read block size of the low level device. This can be independent of
  // the write block size. The device will set the block size to the default
  // block size of 512 bytes.
  void setBlockSize(DevHandle_t handle, int numBytes);

  // The read block size of the low level device. This can be independent of
  // the write block size.
  int getBlockSize(DevHandle_t handle);

  // Print the contents of alls the DevCaps in the DevCaps class.
  void print();

  protected:

  private:

  // This function makes sure that the requested device handle exists. If
  // it does not, a DevCap is inserted in the DevCapMap with the appropriate
  // handle as a key.
  void checkHandle(DevHandle_t devHandle);

  // An STL map used to store DevCaps.
  typedef map<DevHandle_t,DevCap> DevCapsMap;
  DevCapsMap devCaps;

  // The iterator used to find various devices in the DevCapsMap.
  DevCapsMap::iterator it;

  // The input device handle the library's user requests.
  DevHandle_t inputDeviceToUse;

  // The output device handle the library's user requests.
  DevHandle_t outPutDeviceToUse;
};

#endif