// Copyright year by Columbia University. All rights reserved.

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/soundcard.h>

#include <AudioDevices/AudioDeviceException.h>
#include <AudioDevices/OSSDeviceCaps.h>
#include <AudioDevices/OSSAudioDevice.h>
#include <Utils/Debug.h>

// size of the string for an OSS device name
const int OSSAudioDevice::MAX_OSS_NAME_SIZE = 64;

// the max number of audio devices in a linux system
const int OSSAudioDevice::MAX_NUM_OSS_DEVS = 4;

// the max number of audio devices in a linux system
const int OSSAudioDevice::MAX_NUM_OSS_MIXERS = 4;

// stores the probed device names
const string* audioDevNames[OSSAudioDevice::MAX_NUM_OSS_DEVS];

// string used to identify the prefix of the name of the audio device on the system
const string AUDIO_STR = "/dev/audio";

// string used to identify the prefix of the name of the mixer device on the system
const string MIXER_STR = "/dev/mixer";

DeviceCaps*
OSSAudioDevice::getDevCaps()
{
  Debug D("OSSAudioDevice::getDevCaps");

  DeviceCaps *pDeviceCaps = new DeviceCaps();

  for(int audioDev = 0;audioDev < MAX_NUM_OSS_DEVS;audioDev++)
  {
    D.Dbg("audioDev ", audioDev);

    if(getAudioCaps(audioDev,pDeviceCaps) != true)
    {
      D.Dbg("could not get audio device capabilities for ", audioDev);
    }
  }

  for(int mixerDev = 0;mixerDev < MAX_NUM_OSS_MIXERS;mixerDev++)
  {
    D.Dbg("mixerDev ", mixerDev);

    if(getMixerCaps(mixerDev,pDeviceCaps) != true)
    {
      D.Dbg("could not get mixer device capabilities ", mixerDev);
    }
  }

}

// This function collects together information about various devices, so
// it can be presented in a NAL way.
bool 
OSSAudioDevice::getAudioCaps(DevHandle_t audioDev, DeviceCaps *pDeviceCaps)
{
  Debug D("OSSAudioDevice::getAudioCaps");

  // Device numbers seems to start at one, but there is still a /dev/audio, so
  // it needs to be probed as well. The assumption is that the Linux
  // distibution is not too old, so /dev will follow the regular Linux standards
  // for device naming.

  // the file system /dev name
  string* pDeviceName; 
  if(audioDev == 0)
  {
    pDeviceName = new string(AUDIO_STR);
  }
  else
  {
    char *pTmpDevName = new char[MAX_OSS_NAME_SIZE];
    sprintf(pTmpDevName,"%s%d",AUDIO_STR.c_str(),audioDev);
    pDeviceName = new string(pTmpDevName);
  }

  D.Dbg(pDeviceName->c_str(), "");

  int audioDesc;
  audioDesc = ::open(pDeviceName->c_str(), O_RDWR);
  audioDevNames[audioDev] = pDeviceName;

  if(audioDesc < 0)
  {
    D.Dbg("could not find audio device " , audioDev);
    return false;
  }

  pDeviceCaps->setDevName(audioDev,pDeviceName->c_str());

  AudioFormats *audioFormats = new AudioFormats();

  // Check which audio encodings are supported. This only reports what the
  // hardware acutually supports. No driver level sofware encodings are
  // reported. This is probably not a bad thing because we want decent
  // performance. Prior to version 3.6 of OSS MU_LAW is always reported as
  // being supported.
  // Check which sampling modes are supported.
  FormatHandle_t format;
  int rate;
  for(format = 0,rate = 0;rate < NUM_SAMPLE_RATES;
      format++,rate++)
  {
    int sampleRate = rate;
    if(ioctl(audioDesc, SNDCTL_DSP_SPEED, &sampleRate) == -1)
    {
      D.Dbg("sample rate is not supported ", (int)sampleRates[rate]);
    }
    else
    {
      int chan = AudioFormat::MONO;
      if(ioctl(audioDesc, SNDCTL_DSP_STEREO, &chan) == -1)
      {
        D.Dbg("MONO not supported", "");
      }
      else
      {
        AudioFormat *audioFormat = new AudioFormat();	
        audioFormat->setSampleRate(sampleRates[rate]);
        audioFormat->setChannel(AudioFormat::MONO);
        setEncodings(audioDesc,format,audioFormats);
      }
      chan = AudioFormat::STEREO;
      if(ioctl(audioDesc, SNDCTL_DSP_STEREO, &chan) == -1)
      {
        D.Dbg("STEREO not supported","");
      }
      else
      {
        AudioFormat *audioFormat = new AudioFormat();
        audioFormat->setSampleRate(sampleRates[rate]);
        audioFormat->setChannel(AudioFormat::MONO);
        setEncodings(audioDesc,format,audioFormats);
      }
    }
  }

  pDeviceCaps->setAudioFormats(audioDev, audioFormats);
  ::close(audioDesc);
  return true;
}

void
OSSAudioDevice::setEncodings(int audioDesc, FormatHandle_t format,
                             AudioFormats *audioFormats)
{
  Debug D("OSSAudioDevice::setEncodings");

  int mask;
  if(ioctl(audioDesc, SNDCTL_DSP_GETFMTS, &mask) == -1)
  {
    D.Dbg("could not get encoding mask","");
  }
  else
  {
    int encoding;
    for(encoding = 0;encoding < NUM_ENCODINGS;encoding++)
    {
      if(mask & NALEncoding2OSS(encodings[encoding]))
      {
        audioFormats->setEncoding(format, encodings[encoding]);
      }
    }
  }
}

int
OSSAudioDevice::NALEncoding2OSS(AudioFormat::Encoding encoding)
{
  Debug D("OSSAudioDevice::NALEncoding2OSS");

  switch(encoding)
  {
    case AudioFormat::PCMU: return AFMT_MU_LAW;
    case AudioFormat::PCMA: return AFMT_A_LAW;
    case AudioFormat::S8:   return AFMT_S8;
    case AudioFormat::U8:   return AFMT_U8;
    case AudioFormat::S16:  return AFMT_S16_NE;
  }

  return AudioFormat::NO_FORMAT;
}

// The mixers are only probed to make sure some needed capabilities exist. Only
// the name of the device is saved (the name on the filesystem).
bool
OSSAudioDevice::getMixerCaps(DevHandle_t mixerNum, DeviceCaps *deviceCaps)
{
  Debug D("OSSAudioDevice::getMixerCaps");

  // Device numbers seems to start at one, but there is still a /dev/mixer, so
  // it needs to be probed as well. The assumption is that the Linux
  // distibution is not too old, so /dev will follow the regular Linux standards
  // for device naming.
  string *pDeviceName; 
  if(mixerNum == 0)
  {
    pDeviceName = new string(MIXER_STR);
  }
  else
  {
    char *tmp = new char[MAX_OSS_NAME_SIZE];
    sprintf(tmp,"%s%d",MIXER_STR.c_str(),mixerNum);
    pDeviceName = new string(tmp);
  }

  int mixerDesc;
  mixerDesc = ::open(pDeviceName->c_str(), O_RDWR);

  if(mixerDesc < 0)
  {
    // AUDIT The mixer mixerNum was not found.
    return false;
  }

  mixer_info mixerInfo;
  string *pOSSName;
  if(ioctl(mixerDesc, SOUND_MIXER_INFO, &mixerInfo) == -1)
  {
    // AUDIT Could not get mixer info.
  }
  else
  {
    pOSSName = new string(mixerInfo.name);
    // should the name really be the mixer name??
    deviceCaps->setDevName(mixerNum, pOSSName->c_str());
  }

  int mask;
  if(ioctl(mixerDesc, SOUND_MIXER_READ_DEVMASK, &mask) == -1)
  {
    if(mask & SOUND_MIXER_PCM)
    {
      deviceCaps->setInVolSupp(mixerNum, true);
    }
    else
    {
      deviceCaps->setInVolSupp(mixerNum, false);
    }

    if(mask & SOUND_MIXER_MIC)
    {
      deviceCaps->setInVolSupp(mixerNum, true);
    }
    else
    {
      deviceCaps->setInVolSupp(mixerNum, false);
    }
  }
  else
  {
    D.Dbg("could not get device mask from mixer", "");
  }

  ::close(mixerDesc);
  return true;
}

void
OSSAudioDevice::open(DeviceCaps *pDeviceCaps)
{
  Debug D("OSSAudioDevice::open");

  // This pointer should be used in this function, so the
  // OSSDeviceCaps is used and not the DeviceCaps.
  OSSDeviceCaps *pOSSDeviceCaps = (OSSDeviceCaps*)pDeviceCaps;

  inputDeviceToUse = pDeviceCaps->getInputDeviceToUse();
  outputDeviceToUse = pDeviceCaps->getOutputDeviceToUse();

  D.Dbg("device name",
        pOSSDeviceCaps->getAudioDevName(inputDeviceToUse)->c_str());

  audioDesc = ::open(pOSSDeviceCaps->getAudioDevName(inputDeviceToUse)->c_str(),
                     O_RDWR | O_NDELAY);

  if(audioDesc == 0)
  {
    D.Dbg("could not open device", audioDesc);
  }
 
  // This has to be done first.
  if(ioctl(audioDesc, SNDCTL_DSP_SETDUPLEX, 0) == -1)
  {
    D.Dbg("could not set device to duplex", audioDesc);
    ::close(audioDesc);
  }

  // Sampling parameters must be set before the blocksize can be queried.
  int format = NALEncoding2OSS(pOSSDeviceCaps->getEncodingToUse(inputDeviceToUse));
  if(ioctl(audioDesc, SNDCTL_DSP_SETFMT, &format) == -1)
  {
    D.Dbg("the device does not support requested format", audioDesc);
  }

  int rate = pOSSDeviceCaps->getSampleRateToUse(inputDeviceToUse);
  if(ioctl(audioDesc, SNDCTL_DSP_SPEED, &rate) == -1)
  {
    D.Dbg("the device does not support requested sampling rate", audioDesc);
  }

  int channel = pOSSDeviceCaps->getChannelToUse(inputDeviceToUse);
  if(ioctl(audioDesc, SNDCTL_DSP_STEREO, &rate) == -1)
  {
    D.Dbg("the device does not support requested sampling rate", audioDesc);
  }

  // The device automatically computes the fragment size (blocksize) after
  // sampling parameters have been set. For now the blocksize will be
  // calculated this way.
  int fragSize;
  if(ioctl(audioDesc,SNDCTL_DSP_GETBLKSIZE, &fragSize) == -1)
  {
    D.Dbg("could not get the fragment size", audioDesc);
  }

}

bool 
OSSAudioDevice::isOpen()
{
  Debug D("OSSAudioDevice::isOpen");

  return deviceOpen;
}

void
OSSAudioDevice::close()
{
  Debug D("OSSAudioDevice::close");

  ::close(audioDesc);
  deviceOpen = false;
}

int
OSSAudioDevice::read(sample_t *pAppBuff, int buffSize)
{
  audio_buf_info bufInfo;
  ioctl(audioDesc, SNDCTL_DSP_GETISPACE, &bufInfo);
  int bytesAvaiable = min(bufInfo.bytes, buffSize);
  int bytesRead = ::read(audioDesc, (char*)pAppBuff, bytesAvaiable);

  return bytesRead;

}

int
OSSAudioDevice::write(sample_t *pAppBuff, int buffSize)
{
  char *pBuff = (char*)pAppBuff;
  int thisWrite = 0, toWrite = buffSize, written = 0;
  for(; written <= buffSize; written += thisWrite, pBuff += thisWrite,
      toWrite -= thisWrite)
  {
    thisWrite = ::write(audioDesc,pBuff,toWrite);
  }

  return written;
}

// ioctl should be done on the /dev/audio device. There is no way to correlate
// mixers and audio devices, so all the volume functions use the audio device
// file descriptor.

// It is assuemd that only microphone line needs to be adjusted.
void
OSSAudioDevice::setInputVol(int vol)
{
  Debug D("OSSAudioDevice::setInputVol");

  int volume = vol2Dev(vol); 
  if(ioctl(audioDesc,MIXER_WRITE(SOUND_MIXER_MIC),&volume) == -1)
  {
    // AUDIT Could not set input volume.
  }
}

int
OSSAudioDevice::getInputVol()
{
  Debug D("OSSAudioDevice::getInputVol");

  int volume;
  if(ioctl(audioDesc,MIXER_READ(SOUND_MIXER_MIC),&volume) == -1)
  {
    // AUDIT Could not set input volume.
  }

  return volume & 0xff;
}

void
OSSAudioDevice::setOutputVol(int vol)
{
  Debug D("OSSAudioDevice::setOutputVol");

  int volume = vol2Dev(vol); 
  if(ioctl(audioDesc,MIXER_READ(SOUND_MIXER_PCM),&volume) == -1)
  {
    // AUDIT Could not set input volume.
  }
}

int
OSSAudioDevice::getOutputVol()
{
  Debug D("OSSAudioDevice::getOutputVol");

  int volume;
  if(ioctl(audioDesc,MIXER_READ(SOUND_MIXER_MIC),&volume) == -1)
  {
    // AUDIT Could not set input volume.
  }

  return volume & 0x000000ff; // Extract the left volume.
}

int
OSSAudioDevice::vol2Dev(int vol)
{
  Debug D("OSSAudioDevice::vol2Dev");

  int volume = (vol << 8) | vol;
  return volume;
}


int
OSSAudioDevice::getReadBlockSize()
{
  Debug D("OSSAudioDevice::getReadBlockSize");

  return readBlockSize;
} // OSSAudioDevice::getReadBlockSize


int
OSSAudioDevice::getWriteBlockSize()
{
  Debug D("OSSAudioDevice::getReadBlockSize");

  return writeBlockSize;
} // OSSAudioDevice::getWriteBlockSize
