// Copyright 2001 year by Columbia University. All rights reserved.

// Each platform supported by NAL needs to have a class that implements the
// interface between NAL and the lower level platform's device abstraction.
// This class implements the win32 device layer for NAL.

#ifndef OSSAUDIODEVICE_H
#define OSSAUDIODEVICE_H

#include <string>
#include <Common/AudioFormat.h>
#include <AudioDevices/DeviceCaps.h>
#include <AudioDevices/AudioDevice.h>

using namespace std;

// This can be used to loop through the sampleRates.
static AudioFormat::SampleRate sampleRates[NUM_SAMPLE_RATES] = {
  AudioFormat::S8K,
  AudioFormat::S11K,
  AudioFormat::S16K,
  AudioFormat::S22K,
  AudioFormat::S24K,
  AudioFormat::S32K,
  AudioFormat::S44K,
  AudioFormat::S48K
};

// This can be used to loop through the encodings.
static AudioFormat::Encoding encodings[NUM_ENCODINGS] = {
  AudioFormat::PCM,
  AudioFormat::PCMU,
  AudioFormat::PCMA,
  AudioFormat::S8,
  AudioFormat::U8,
  AudioFormat::S16
};

class OSSAudioDevice : public AudioDevice
{
  public:

  static const int MAX_NUM_OSS_DEVS;
  static const int MAX_NUM_OSS_MIXERS;
  static const int MAX_OSS_NAME_SIZE;

  // Since only the lower level device can correctly fill out a device
  // capability class, this function is not inherited from AudioDevice. 
  // The caller of this function is expected to use the DeviceCaps class to
  // find the capabilities of the low level device and then choose how to
  // initialize that device from the capabilities described in the DeviceCaps
  // class.
  static DeviceCaps* getDevCaps();

  OSSAudioDevice(){};
  ~OSSAudioDevice(){};

  // The funciton getDevCaps() on the low level driver should be called
  // to figure out what the device can support before trying to open the
  // device. If the device does not support the requested audioFormat, the
  // device will not open correctly. The open function will throw an
  // AudioDeviceException, if the device could not be open successfully.
  void open(DeviceCaps *pDeviceCaps);

  // This funtion returns true if the device is open and false if the
  // device is not open.
  bool isOpen();

  // Closes the device. The close function will throw an AudioDeviceException,
  // if the device could not be closed successfully.
  void close();

  // The read function reads data from the audio device's lower level
  // device. The buffer does not have to be of the same size as the
  // block size. The first argument is the buffer to be filled and
  // the second argument is the size of the buffer. The type of data
  // the buffer is filled with depends on the information the device
  // was open with using DeviceCaps.
  // buffSize is in bytes and the number of bytes read is returned.
  int read(sample_t *pAppBuff, int buffSize);

  // The write function writes data to the audio device's lower level
  // device. The buffer does not have to be of the same size as the
  // block size. The first argument is the buffer to be filled and
  // the second argument is the size of the buffer. The type of data
  // the buffer is filled with depends on the information the device
  // was open with using DeviceCaps. 
  // buffSize is in bytes and the number of bytes read is returned.
  int write(sample_t *pAppBuff, int buffSize);

  // Sets the volume on the input (mirophone) device. Volume can be 0-100.
  void setInputVol(int vol);

  // Returns the input (mirophone) device volume current setting. Volume
  // can be 0-100.
  int getInputVol();


  void setOutputVol(int vol);
  int getOutputVol();

  // Returns the present sample rate. Can be called any time after the
  // device has been opened and is still open.
  AudioFormat::Channel getNumChannels();

  // Returns the present number of channels. Can be called any time after
  // the device has been opened and is still open.
  AudioFormat::SampleRate getSampleRate();

  // Returns the present size of the read blocks being used by the device.
  // Can be called any time after the device has been opened and is still open.
  int getReadBlockSize();

  // Returns the present size of the read blocks being used by the device.
  // Can be called any time after the device has been opened and is still open.
  int getWriteBlockSize();
 
  protected:

  private:

  // Converts NAL encodings to OSS wave encodings.
  static int NALEncoding2OSS(AudioFormat::Encoding encoding);

  // It may be necessary to do some convertion of volumes so the lower level
  // device understands the 0-100 volume scale of the NAL.
  static int vol2Dev(int vol);

  // audio device    
  static const string AUDIO_STR;

  // the name the OSS gives the device
  string audioName;

  // The descriptor returned by the open function. The same descriptor is used
  // for mixer ioctl's.
  int audioDesc;

  int readBuffSize; // in bytes
  int writeBuffSize; // in bytes
  int writeBlockSize; // in bytes
  int readBlockSize; // in bytes
  static string* audioDevNames[];
  
  // mixer device
  static const string MIXER_STR;
  string mixerName; // the name the OSS gives the mixer
  string mixerDevName; // the file system /dev name
  int readDevMask;
  int readRecMask;

  // Gets the capabilities of the audio device and puts them in the
  // DevCaps class handed in.
  static bool getAudioCaps(DevHandle_t audioNum, DeviceCaps *pDeviceCaps);

  // Sets the supported encodings in the AudioFormats class handed in.
  static void setEncodings(int audioDesc, FormatHandle_t format,
                           AudioFormats *audioFormats);

  // Gets the capabilities of the mixer device and puts them in the
  // DevCaps class handed in.
  static bool getMixerCaps(DevHandle_t mixerNum, DeviceCaps *pDeviceCaps);
};

#endif
