// Copyright 2001 by Columbia University. All rights reserved.

// Each platform supported by NAL needs to have a class that implements the
// interface between NAL and the lower level platform's device abstraction.
// This class implements the win32 device layer for NAL.

#ifndef WIN32AUDIODEVICE_H
#define WIN32AUDIODEVICE_H

#include <windows.h>
#include <mmsystem.h>
#include <mmreg.h>
#include <AudioDevices/AudioDevice.h>
#include <AudioDevices/Win32DeviceCaps.h>
#include <Utils/Mutex.h>

using namespace std;

// Used to loop through the sampleRates.
static AudioFormat::SampleRate sampleRates[NUM_SAMPLE_RATES] = {
  AudioFormat::S8K,
  AudioFormat::S11K,
  AudioFormat::S16K,
  AudioFormat::S22K,
  AudioFormat::S24K,
  AudioFormat::S32K,
  AudioFormat::S44K,
  AudioFormat::S48K
};

// Used to loop through the encodings.
static AudioFormat::Encoding encodings[NUM_ENCODINGS] = {
  AudioFormat::PCM,
  AudioFormat::PCMU,
  AudioFormat::PCMA,
  AudioFormat::S8,
  AudioFormat::U8,
  AudioFormat::S16
};

// The number of win32 wave in devices in the system
static UINT numInWaves;

// The number of win32 wave out devices in the system
static UINT numOutWaves;

// The number of win32 mixers the system
static UINT numMixers;

// Used by the static functions in this class to fill out the DeviceCaps.
static Win32DeviceCaps *pWin32DeviceCaps;

class Win32AudioDevice : public AudioDevice
{
  public:

  static const int MAX_WAVE_VOLUME;

  // Since only the lower level device can correctly fill out a device
  // capability class, this function is not inherited from AudioDevice. 
  // The caller of this function is expected to use the DeviceCaps class to
  // find the capabilities of the low level device and then choose how to
  // initialize that device from the capabilities described in the DeviceCaps
  // class.
  static DeviceCaps* getDevCaps();

  Win32AudioDevice();
  ~Win32AudioDevice();

  // The funciton getDevCaps() on the low level driver should be called
  // to figure out what the device can support before trying to open the
  // device. If the device does not support the requested audioFormat, the
  // device will not open correctly. The open function will throw an
  // AudioDeviceException, if the device could not be open successfully.
  void open(DeviceCaps *pDeviceCaps);

  // This funtion returns true if the device is open and false if the
  // device is not open.
  bool isOpen();

  // Closes the device. The close function will throw an AudioDeviceException,
  // if the device could not be closed successfully.
  void close();

  // The read function reads data from the audio device's lower level
  // device. The buffer does not have to be of the same size as the
  // block size. The first argument is the buffer to be filled and
  // the second argument is the size of the buffer. The type of data
  // the buffer is filled with depends on the information the device
  // was open with using DeviceCaps.
  // buffSize is in bytes and the number of bytes read is returned.
  int read(sample_t *pAppBuff, int buffSize);

  // The write function writes data to the audio device's lower level
  // device. The buffer does not have to be of the same size as the
  // block size. The first argument is the buffer to be filled and
  // the second argument is the size of the buffer. The type of data
  // the buffer is filled with depends on the information the device
  // was open with using DeviceCaps. 
  // buffSize is in bytes and the number of bytes read is returned.
  int write(sample_t *pAppBuff, int buffSize);

 // Sets the volume on the input (mirophone) device. Volume can be 0-100.
  void setInputVol(int vol);

  // Returns the input (mirophone) device volume current setting. Volume
  // can be 0-100.
  int getInputVol();

  // Sets the volume on the output (speakers) device. Volume can be 0-100.
  void setOutputVol(int vol);

  // Returns the output (speakers) device volume current setting. Volume
  // can be 0-100.
  int getOutputVol();

  // Returns the present sample rate. Can be called any time after the
  // device has been opened and is still open.
  AudioFormat::Channel getNumChannels();

  // Returns the present number of channels. Can be called any time after
  // the device has been opened and is still open.
  AudioFormat::SampleRate getSampleRate();

  // Returns the present size of the read blocks being used by the device.
  // Can be called any time after the device has been opened and is still open.
  int getReadBlockSize();

  // Returns the present size of the read blocks being used by the device.
  // Can be called any time after the device has been opened and is still open.
  int getWriteBlockSize();
  
  protected:

  private:

  // Since the driver has 2 threads, this mutex is used to protect the read
  // data structures.
  Mutex *pMutex;

  // The AudioFormat presently being set by the device.
  static int audioFormatKey;

  // The number of wave headers (blocks) used in the device's internal
  // read or write list.
  const static int NUM_BLOCKS;

  // The default size of one block used in a list of wave headers.
  const static int DEFAULT_BLOCK_SIZE;

  // Functions to get audio device information. These are all static,
  // so they can be called by the getDevCaps or eachother.

  // Gets the capabilities of the wave in device and puts them in the
  // DevCaps class handed in.
  static bool getWaveInCaps(int waveInDev,int audioDev,
                            Win32DeviceCaps *pDevCaps);

  // Gets the capabilities of the wave out device and put them in the
  // DevCaps class handed in.
  static bool getWaveOutCaps(int waveOutDev,int audioDev,
                             Win32DeviceCaps *pDevCaps);

  // Sets the wave in encodings in the AudioFormats class handed in.
  static void setInEncodings(int waveIn,
                             FormatHandle_t handle,
                             AudioFormats *pAudioFormats,
                             AudioFormat::SampleRate rate,
                             AudioFormat::Channel channels,
                             AudioFormat::NumBits bits);

  // Sets the wave in encodings in the AudioFormats class handed in.
  static void setOutEncodings(int waveOut,
                              FormatHandle_t handle,
                              AudioFormats *pAudioFormats,
                              AudioFormat::SampleRate rate,
                              AudioFormat::Channel channels,
                              AudioFormat::NumBits bits);

  // Converts NAL encodings to win32 wave encodings.
  static int NAL2WaveEncoding(AudioFormat::Encoding encoding);

  // Returns true if the sample rate, channel, and numbits is supported
  // by the device in question. Returns false if the combination is not
  // supported.
  static bool NAL2WaveFormat(DWORD dwFormat,AudioFormat::SampleRate rate,
                             AudioFormat::Channel channel,
                             AudioFormat::NumBits bits);

  // For error messages from win32 calls.
  char errorText[MAXERRORLENGTH];

  // It may be necessary to do some convertion of volumes so the lower level
  // device understands the 0-100 volume scale of the NAL.
  int vol2Dev(int vol);
  int dev2Vol(int vol);

  // true when the wave in device is closed, false when the device is open.
  bool waveInClosed;

  // handle to the wave in device
  HWAVEIN waveInHandle;
  
  // Used to point to the chuck of memory used for the wave in list.
  WAVEHDR *pWaveInHdrs;

  // Head of the wave in header list used by waveInProc
  WAVEHDR *pFirstWaveHdr;

  // buffer used for reading from wave device (microphone)
  char *pWaveInBuf;

  // Opens the wave in device and sets up the wave in buffers.
  // Throws an AudioDeviceException if a problem occurs.
  void openWaveIn();

  // Closes the wave in device and reset all the pointers and buffers.
  void closeWaveIn();

  // A handle to the read thread used by waveInProc function.
  HANDLE readThread;

  // The ID of the read thread.
  DWORD readThreadID;

  // true when the wave out device is closed, false when the device is open.
  bool waveOutClosed;

  // A handle to the wave out device;
  HWAVEOUT waveOutHandle;

  // Used to point to the chuck of memory used for the wave out list.
  WAVEHDR *pWaveOutHdrs;

  // Buffer used for write to wave device (speakers).
  char *pWaveOutBuf;

  // Finds the next available wave header to put in the write buffer.
  WAVEHDR* getOutHdr();

  // Opens the wave out device and sets up the wave out buffers.
  // Throws an AudioDeviceException if a problem occurs.
  void openWaveOut();

  // Closes the wave in device and reset all the pointers and buffers.
  void closeWaveOut();

  // This static function knows how to call the function on a class
  // instance.
  static void readProc(Win32AudioDevice *pWin32Dev);

  // This function is the function called on the instance.
  void waveInProc();

  // True if the mixer is closed and false if it is open.
  bool mixerClosed;

  // Handle to the open mixer.
  HMIXER mixerHandle;

  // Open the mixer.
  void openMixer();

  // Close the mixer.
  void closeMixer();

  // Print the statics of one wave header.
  void printHdrStats(WAVEHDR *pWaveHdrs);
};

#endif