// Codec.h: Definition of Codec class.  Abstract representation
// of the formats provided by a codec.  Provides common interface 
// for all codecs.  Codecs are created using the CodecFactory.
// 
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////

#ifndef _CODEC_H_
#define _CODEC_H_

#include <list>
#include <Common/AudioFormat.h>
#include <Common/CodecFormat.h>

using namespace std;

typedef list<CodecFormat> CodecFormatList;

class CodecState;

class Codec
{
public:
  //Tells the codec to reset its state.
  //This should be called before begining processing on
  //a new audio stream.
	virtual void resetState() =0;

  //Converts the input buffer from raw sound bytes into
  //encoded sound bytes.  Before calling encode, the current
  //codec format must be set.
  //@param inbuf array of 16 bit PCM data should be equal to the length
  //of AudioFormat::bytesPerBlock / sizeof(short)
  //@param outbuf array of chars which upon return will contain the encoded
  //data. This should be as long as CodecFormat::framesize.
  //@param length The length of the oubuf that was used is returned in bytes.
  //This is typically eqaul to the CodecFormat::framesize member.
	virtual void 
    encode(unsigned short *inbuf, unsigned char *outbuf, int &length) = 0;

  //Converts the input buffer from encodec sound bytes into raw sound bytes.
  //Before calling decode the current format must be set.
  //@param inbuf array of chars which upon return will contain the encoded
  //data. This should be as long as CodecFormat::framesize.
  //@param outbufarray of 16 bit PCM data should be equal to the length
  //of AudioFormat::bytesPerBlock / sizeof(short)
  //@param length The length of the oubuf that was used is returned.
  //Typically equal to AudioFormat::bytesPerBlock / sizeof(short)
	virtual void 
    decode(unsigned char *inbuf, unsigned short *outbuf, int &length) = 0;

  //Sets the current codec format
  //@param the codec format to be made current
	inline virtual void setCodecFormat(CodecFormat &codecFormat){
		delete this->codecFormat;
		this->codecFormat = new CodecFormat(codecFormat);
	}

  //returns the currently set coded format
  //@return Current codec format.
  inline virtual CodecFormat *getCodecFormat(){
    return codecFormat;
  }

  //Gets the list of formats that this codec supports.
  //@return STL list object that contains CodecFormat objects.
  virtual const CodecFormatList &getSupportedFormats(){
    return formatlist;
  }

  //Creates a new Codec instance.  Initializes 
  //pointers to NULL
	Codec():codecFormat(NULL),codecState(NULL){}

  //Destructor
	~Codec();
protected:
  //pointer to current codec format.  Allocated by
  //Codec::setCodecFormat, deleted in Codec destructor.
	CodecFormat *codecFormat;

  //pointer to decode/encode state information.  This is allocated
  //on an as needed basis by Codec subclasses.  It is deleted in 
  //the Codec destructor.
	CodecState *codecState;

  //A STL list that contains CodecFormat objects.  Populated
  //by Codec subclass' constructors.
  CodecFormatList formatlist;
};

#endif
