// g711.cpp: Implementation of G711 Codec class
// 
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////

#ifdef _MSC_VER
#pragma warning(disable:4786) //disables nasty vc++ warning caused by stl
#endif

#include "g711.h"
#include <Utils/Debug.h>

short G711::seg_aend[TABLE_SIZE] = {0x1F, 0x3F, 0x7F, 0xFF,
0x1FF, 0x3FF, 0x7FF, 0xFFF};

short G711::seg_uend[TABLE_SIZE] = {0x3F, 0x7F, 0xFF, 0x1FF,
0x3FF, 0x7FF, 0xFFF, 0x1FFF};

/* copy from CCITT G.711 specifications */

/* u- to A-law conversions */
unsigned char G711::u2a[128] = {			
  1,	1,	2,	2,	3,	3,	4,	4,
    5,	5,	6,	6,	7,	7,	8,	8,
    9,	10,	11,	12,	13,	14,	15,	16,
    17,	18,	19,	20,	21,	22,	23,	24,
    25,	27,	29,	31,	33,	34,	35,	36,
    37,	38,	39,	40,	41,	42,	43,	44,
    46,	48,	49,	50,	51,	52,	53,	54,
    55,	56,	57,	58,	59,	60,	61,	62,
    64,	65,	66,	67,	68,	69,	70,	71,
    72,	73,	74,	75,	76,	77,	78,	79,
    80,	82,	83,	84,	85,	86,	87,	88,
    89,	90,	91,	92,	93,	94,	95,	96,
    97, 98, 99, 100, 101, 102, 103, 104,
    105, 106, 107, 108, 109, 110, 111, 112,
    113, 114, 115, 116, 117, 118, 119, 120,
    121, 122, 123, 124, 125, 126, 127, 128
};

/* A- to u-law conversions */
unsigned char G711::a2u[128] = {
  1,	3,	5,	7,	9,	11,	13,	15,
    16,	17,	18,	19,	20,	21,	22,	23,
    24,	25,	26,	27,	28,	29,	30,	31,
    32,	32,	33,	33,	34,	34,	35,	35,
    36,	37,	38,	39,	40,	41,	42,	43,
    44,	45,	46,	47,	48,	48,	49,	49,
    50,	51,	52,	53,	54,	55,	56,	57,
    58,	59,	60,	61,	62,	63,	64,	64,
    65,	66,	67,	68,	69,	70,	71,	72,
    73,	74,	75,	76,	77,	78,	79,	80,
    80,	81,	82,	83,	84,	85,	86,	87,
    88,	89,	90,	91,	92,	93,	94,	95,
    96,	97,	98,	99,	100, 101, 102, 103,
    104, 105, 106, 107, 108, 109, 110, 111,
    112, 113, 114, 115, 116, 117, 118, 119,
    120, 121, 122, 123, 124, 125, 126, 127
};

// G711::linear2alaw(short pcm_val)
// See g711.h for function comment.
unsigned char G711::linear2alaw(short pcm_val)
{
  short	 mask;
  short	 seg;
  unsigned char aval;

  //Debug D("G711::linear2alaw");
  
  pcm_val = pcm_val >> 3;
  
  if (pcm_val >= 0) {
    mask = 0xD5;		/* sign (7th) bit = 1 */
  } else {
    mask = 0x55;		/* sign bit = 0 */
    pcm_val = -pcm_val - 1;
  }
  
  /* Convert the scaled magnitude to segment number. */
  seg = searchAEnd(pcm_val);
  
  /* Combine the sign, segment, and quantization bits. */
  
  if (seg >= 8)		/* out of range, return maximum value. */
    return (unsigned char) (0x7F ^ mask);
  else {
    aval = (unsigned char) seg << SEG_SHIFT;
    if (seg < 2)
      aval |= (pcm_val >> 1) & QUANT_MASK;
    else
      aval |= (pcm_val >> seg) & QUANT_MASK;
    return (aval ^ mask);
  }
}// G711::linear2alaw(short pcm_val)

// G711::alaw2linear(unsigned char	a_val)
// See g711.h for function comment
short G711::alaw2linear(unsigned char	a_val)
{
  short t;
  short seg;

  //Debug D("G711::alaw2linear");
  
  a_val ^= 0x55;
  
  t = (a_val & QUANT_MASK) << 4;
  seg = ((unsigned)a_val & SEG_MASK) >> SEG_SHIFT;
  switch (seg) {
  case 0:
    t += 8;
    break;
  case 1:
    t += 0x108;
    break;
  default:
    t += 0x108;
    t <<= seg - 1;
  }
  return ((a_val & SIGN_BIT) ? t : -t);
}//G711::alaw2linear(unsigned char	a_val)

// G711::linear2ulaw(short pcm_val)
// See g711.h for function comment
unsigned char G711::linear2ulaw(short pcm_val)
{
  short         mask;
  short	 seg;
  unsigned char uval;

  //Debug D("G711::linear2ulaw");
  
  /* Get the sign and the magnitude of the value. */
  pcm_val = pcm_val >> 2;
  if (pcm_val < 0) {
    pcm_val = -pcm_val;
    mask = 0x7F;
  } else {
    mask = 0xFF;
  }
  if ( pcm_val > CLIP ) pcm_val = CLIP;		/* clip the magnitude */
  pcm_val += (BIAS >> 2);
  
  /* Convert the scaled magnitude to segment number. */
  seg = searchUEnd(pcm_val);
  
  /*
  * Combine the sign, segment, quantization bits;
  * and complement the code word.
  */
  if (seg >= 8)		/* out of range, return maximum value. */
    return (unsigned char) (0x7F ^ mask);
  else {
    uval = (unsigned char) (seg << 4) | ((pcm_val >> (seg + 1)) & 0xF);
    return (uval ^ mask);
  }
}//G711::linear2ulaw(short pcm_val)

// G711::ulaw2linear(unsigned char u_val)
// see g711.h for function comment
short G711::ulaw2linear(unsigned char u_val)
{
  //Debug D("G711::ulaw2linear");

  short t;
  
  /* Complement to obtain normal u-law value. */
  u_val = ~u_val;
  
  /*
  * Extract and bias the quantization bits. Then
  * shift up by the segment number and subtract out the bias.
  */
  t = ((u_val & QUANT_MASK) << 3) + BIAS;
  t <<= ((unsigned)u_val & SEG_MASK) >> SEG_SHIFT;
  
  return ((u_val & SIGN_BIT) ? (BIAS - t) : (t - BIAS));
}//G711::ulaw2linear(unsigned char u_val)

// G711::G711()
// Constructor.  Initializes g711 codec formats and tables needed
// to convert from linear pcm to g711 and back.
G711::G711()
{
  //initialze supported formats
  /* 8kHz -law */
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-8K-Mono",
    "ITU G.711 -law codec.  Sun Microsystems public implementation.",
    0,0,160,
    AudioFormat(AudioFormat::S16,AudioFormat::S8K,AudioFormat::B16,
    AudioFormat::MONO,
				G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-8K-Stereo",
    "ITU G.711 -law codec.  Sun Microsystems public implementation.",
    83,0,320,
    AudioFormat(AudioFormat::S16,AudioFormat::S8K,AudioFormat::B16,
    AudioFormat::STEREO,
				2 * G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  /* 16kHz -law */
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-16K-Mono",
    "ITU G.711 -law codec.  Sun Microsystems public implementation.",
    84,0,160,
    AudioFormat(AudioFormat::S16,AudioFormat::S16K,AudioFormat::B16,
    AudioFormat::MONO,
				G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-16K-Stereo",
    "ITU G.711 -law codec.  Sun Microsystems public implementation.",
    85,0,320,
    AudioFormat(AudioFormat::S16,AudioFormat::S16K,AudioFormat::B16,AudioFormat::STEREO,
				2 * G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  /* 32kHz -law */
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-32K-Mono",
    "ITU G.711 -law codec.  Sun Microsystems public implementation.",
    86,0,160,
    AudioFormat(AudioFormat::S16,AudioFormat::S32K,AudioFormat::B16,
    AudioFormat::MONO,
				G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-32K-Stereo",
    "ITU G.711 -law codec.  Sun Microsystems public implementation.",
    87,0,320,
    AudioFormat(AudioFormat::S16,AudioFormat::S32K,AudioFormat::B16,
    AudioFormat::STEREO,
				2 * G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  /* 48kHz -law */
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-48K-Mono",
    "ITU G.711 -law codec.  Sun Microsystems public implementation.",
    88,0,160,
    AudioFormat(AudioFormat::S16,AudioFormat::S48K,AudioFormat::B16,
    AudioFormat::MONO,
				G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-48K-Stereo",
    "ITU G.711 -law codec.  Sun Microsystems public implementation.",
    89,0,320,
    AudioFormat(AudioFormat::S16,AudioFormat::S48K,AudioFormat::B16,
    AudioFormat::STEREO,
				2 * G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  /* 8kHz A-law */
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-8K-Mono",
    "ITU G.711 A-law codec.  Sun Microsystems public implementation.",
    8,0,160,
    AudioFormat(AudioFormat::S16,AudioFormat::S8K,AudioFormat::B16,
    AudioFormat::MONO,
				G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-8K-Stereo",
    "ITU G.711 A-law codec.  Sun Microsystems public implementation.",
    90,0,320,
    AudioFormat(AudioFormat::S16,AudioFormat::S8K,AudioFormat::B16,
    AudioFormat::STEREO,
				2 * G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  /* 16kHz A-law */
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-16K-Mono",
    "ITU G.711 A-law codec.  Sun Microsystems public implementation.",
    91,0,160,
    AudioFormat(AudioFormat::S16,AudioFormat::S16K,AudioFormat::B16,
    AudioFormat::MONO,
				G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-16K-Stereo",
    "ITU G.711 A-law codec.  Sun Microsystems public implementation.",
    124,0,320,
    AudioFormat(AudioFormat::S16,AudioFormat::S16K,AudioFormat::B16,AudioFormat::STEREO,
				2 * G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  /* 32kHz A-law */
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-32K-Mono",
    "ITU G.711 A-law codec.  Sun Microsystems public implementation.",
    92,0,160,
    AudioFormat(AudioFormat::S16,AudioFormat::S32K,AudioFormat::B16,AudioFormat::MONO,
				G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-32K-Stereo",
    "ITU G.711 A-law codec.  Sun Microsystems public implementation.",
    93,0,320,
    AudioFormat(AudioFormat::S16,AudioFormat::S32K,AudioFormat::B16,AudioFormat::STEREO,
				2 * G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  /* 48kHz A-law */
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-48K-Mono",
    "ITU G.711 A-law codec.  Sun Microsystems public implementation.",
    94,0,160,
    AudioFormat(AudioFormat::S16,AudioFormat::S48K,AudioFormat::B16,AudioFormat::MONO,
				G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  formatlist.insert(formatlist.end(),
    CodecFormat(
    "PCMU-48K-Stereo",
    "ITU G.711 A-law codec.  Sun Microsystems public implementation.",
    95,0,320,
    AudioFormat(AudioFormat::S16,AudioFormat::S48K,AudioFormat::B16,
    AudioFormat::STEREO,
				2 * G711_SAMPLES_PER_FRAME * NUM_BYTES_PER_SAMPLE
        )
        )
        );
  
  int i;

  //initialize conversion tables
  
  for(i = 0; i < 256; i++)
    mulawtolin[i] = ulaw2linear((unsigned char)i);
  
  for(i = -32767; i < 32768; i++) 
    lintomulaw[(unsigned short)i] = linear2ulaw((short)i);
  
  for(i = 0; i < 256; i++) 
    alawtolin[i] = alaw2linear((unsigned char)i);
  
  for(i = -32767; i < 32768; i+= 8) 
    lintoalaw[(unsigned short)i>>3] = linear2alaw((short)i);
  
  //g711 does not use the codecstate
  codecState = NULL;
  
}//G711::G711()


// G711::encode(unsigned short *inbuf, unsigned char *outbuf, int &length)
// Taken from the publicly available Sun implementation.
void G711::encode(unsigned short *inbuf, unsigned char *outbuf, int &length)
{
  Debug D("G711::encode");
  
  int len;
  unsigned char *p, *pe;
  
  //    assert(idx < G711_NUM_FORMATS);
  //    UNUSED(state);
  
  /* No difference between A-law and U-law sizes */
  //len = cs[idx].mean_coded_frame_size;
  len = codecFormat->getFrameSize();
  
  p  = outbuf;
  pe = outbuf + len;
  
  if(codecFormat->getName().find("PCMU") != string::npos){
    D.Dbg("medium","G711 Encoding: ","MU-Law");
    while(p != pe) {
      *p = lintomulaw[(*inbuf)];
      p++; inbuf++;
    }
  } else {
    D.Dbg("medium","G711 Encoding: ","A-Law");
    while(p != pe) {
      *p = lintoalaw[(*inbuf)];
      p++; inbuf++;
    }
  }
  
  length = len;
}//G711::encode(unsigned short *inbuf, unsigned char *outbuf, int &length)

// G711::decode(unsigned char *inbuf, unsigned short *outbuf, int &length)
// Taken from the publicly available Sun implementation.
void G711::decode(unsigned char *inbuf, unsigned short *outbuf, int &length)
{
  Debug D("G711::decode");
  
  int len;
  unsigned char *p, *pe;
  unsigned short *sp;
  
  len = codecFormat->getFrameSize();
  
  p  = inbuf;
  pe = inbuf + len;

  sp = outbuf;
  
  if(codecFormat->getName().find("PCMU") != string::npos){

    D.Dbg("high","G711 Encoding: ","Mu-Law");

    while(p != pe) {
      *sp = mulawtolin[(*p)];
      sp++; p++;
    }
  } else {

    D.Dbg("high","G711 Encoding: ","A-Law");

    while(p != pe) {
      *sp = alawtolin[(*p)];
      sp++; p++;
    }
  }
  
  length = len;
}//G711::decode(unsigned char *inbuf, unsigned short *outbuf, int &length)

// G711::resetState()
// state not used in g711
void G711::resetState()
{
  //this function intentionally left blank
}
