// g711.h Definition of G711 class.  Implements the
// abstract Codec interface.  Function comments may be found in Codec.h
// 
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////

#if !defined(_G711_H_)
#define _G711_H_

#include "Codec.h"
#include <Utils/factory.h>

#include <iostream>
#include <string>

using namespace std;

const int G711_SAMPLES_PER_FRAME = 160;
const int TABLE_SIZE = 8;
const int BIAS = 0x84;		/* Bias for linear code. */
const int CLIP = 8159;
const int SIGN_BIT = 0x80;	/* Sign bit for a A-law byte. */
const int QUANT_MASK = 0xf;  /* Quantization field mask. */
const int NSEGS = 8;         /* Number of A-law segments. */
const int SEG_SHIFT = 4;     /* Left shift for segment number. */
const int SEG_MASK = 0x70;   /* Segment field mask. */

class G711 : public Codec
{
public: 
  
  //Functions defined by abstract Codec class.  See Codec.h for definition.
	virtual void resetState();
  virtual void 
    decode(unsigned char *inbuf, unsigned short *outbuf, int &length);
  virtual void 
    encode(unsigned short *inbuf, unsigned char *outbuf, int &length);
  
  //Creates a new G711 codec object.
  G711();
  
protected:

  //Specify table to search by calling
  //function instead of parameter.
  //@param value to search for
  //@return The table index where the value was found.
  static short searchAEnd(short val){
    return search(val,seg_aend);
  }

  //Specify table to search by calling
  //function instead of parameter.
  //@param value to search for
  //@return The table index where the value was found.  
  static short searchUEnd(short val){
    return search(val,seg_uend);
  }
  
  //serach for value in table.
  //returns location or TABLE_SIZE if not found
  static short search(short val,short *table){
	   short i;
     
     for (i = 0; i < TABLE_SIZE; i++) {
       if (val <= table[i])
         return (i);
     }
     
     return (TABLE_SIZE);
  }
  
  /*
	 * linear2alaw() - Convert a 16-bit linear PCM value to 8-bit A-law
   *
   * linear2alaw() accepts an 16-bit integer and encodes it as A-law data.
   *
   *		Linear Input Code	Compressed Code
   *	------------------------	---------------
   *	0000000wxyza			000wxyz
   *	0000001wxyza			001wxyz
   *	000001wxyzab			010wxyz
   *	00001wxyzabc			011wxyz
   *	0001wxyzabcd			100wxyz
   *	001wxyzabcde			101wxyz
   *	01wxyzabcdef			110wxyz
   *	1wxyzabcdefg			111wxyz
   *
   * For further information see John C. Bellamy's Digital Telephony, 1982,
   * John Wiley & Sons, pps 98-111 and 472-476.
	 */
  static unsigned char linear2alaw(short pcm_val);
  //alaw2linear() - Convert an A-law value to 16-bit linear PCM
  static short alaw2linear(unsigned char	a_val);
  
  /*
  * linear2ulaw() - Convert a linear PCM value to u-law
  *
  * In order to simplify the encoding process, the original linear magnitude
  * is biased by adding 33 which shifts the encoding range from (0 - 8158) to
  * (33 - 8191). The result can be seen in the following encoding table:
  *
  *	Biased Linear Input Code	Compressed Code
  *	------------------------	---------------
  *	00000001wxyza			000wxyz
  *	0000001wxyzab			001wxyz
  *	000001wxyzabc			010wxyz
  *	00001wxyzabcd			011wxyz
  *	0001wxyzabcde			100wxyz
  *	001wxyzabcdef			101wxyz
  *	01wxyzabcdefg			110wxyz
  *	1wxyzabcdefgh			111wxyz
  *
  * Each biased linear code has a leading 1 which identifies the segment
  * number. The value of the segment number is equal to 7 minus the number
  * of leading 0's. The quantization interval is directly available as the
  * four bits wxyz.  * The trailing bits (a - h) are ignored.
  *
  * Ordinarily the complement of the resulting code word is used for
  * transmission, and so the code word is complemented before it is returned.
  *
  * For further information see John C. Bellamy's Digital Telephony, 1982,
  * John Wiley & Sons, pps 98-111 and 472-476.
  */
  
  // 2's complement (16-bit range)
  static unsigned char linear2ulaw(short pcm_val);
  
  /*
	 * ulaw2linear() - Convert a u-law value to 16-bit linear PCM
   *
   * First, a biased linear code is derived from the code word. An unbiased
   * output can then be obtained by subtracting 33 from the biased code.
   *
   * Note that this function expects to be passed the complement of the
   * original code word. This is in keeping with ISDN conventions.
	 */
  static short ulaw2linear(unsigned char u_val);
  
protected:
   
  //tables for converting values
  static unsigned char u2a[128];
  static unsigned char a2u[128];
  
  static short seg_aend[TABLE_SIZE];
  static short seg_uend[TABLE_SIZE];
  
  short mulawtolin[256];
  unsigned char lintomulaw[65536];
  
  short alawtolin[256];
  unsigned char lintoalaw[8192]; 
  
};

namespace{
  static registerInFactory<Codec, G711> registerG711("G711");
  static registerInFactory<Codec, G711> registerg711("g711");
}

#endif //!defined(_G711_H_)
