// Copyright 2001 year by Columbia University. All rights reserved.

// Information about audio types is stored in this class. The types are:
// sample rate - the sample rate measured in Khz for data.
// encoding - the way the data in a buffer is encoded.
// channel - if source of the data is mono or stereo.
// numbits - the number of bits used for the sample data.

#ifndef AUDIOFORMAT_H_INCLUDED
#define AUDIOFORMAT_H_INCLUDED

#ifdef WIN32
#include <windows.h>
#endif

#include <Utils/Debug.h>

static const int NUM_SAMPLE_RATES = 8;
static const int NUM_ENCODINGS = 6;
static const int NUM_CHANNELS = 2;
static const int NUM_BITS = 2;
static const int NUM_BYTES_PER_SAMPLE = 2;

typedef int FormatHandle_t;

class AudioFormat
{
  public:	  

  const static int NO_FORMAT;

  typedef enum {
    S8K = 8000,
    S11K = 11025,
    S16K = 16000,
    S22K = 22050,
    S24K = 24000,
    S32K = 32000,
    S44K = 44100,
    S48K = 48000
  } SampleRate;

  // The various encoding types which may be supported by the device.
  typedef enum {
    PCM,
    PCMU, // mu-law (8 bits)
    PCMA, // a-law (8 bits)
    S8,   // signed 8 bits
    U8,   // unsigned 8 bits
    S16   // signed 16 bits
  } Encoding;

  typedef enum {
    MONO = 1,
    STEREO = 2
  } Channel;

  typedef enum {
    B8 = 8,
    B16 = 16
  } NumBits;

  AudioFormat();

  AudioFormat(Encoding encoding, SampleRate sampleRate, 
              NumBits numBits, Channel channel, int blocksize)
              :sampleRate(sampleRate), encoding(encoding),
              channel(channel), numBits(numBits), 
              bytesPerBlock(blocksize) {};

  AudioFormat( const AudioFormat &right);

  ~AudioFormat ();

  inline void setSampleRate(SampleRate samplerate);
  inline const SampleRate getSampleRate();

  inline void setEncoding(Encoding encoding);
  inline const Encoding getEncoding();
  
  inline const Channel getChannel();
  inline void setChannel(Channel channel);

  inline void setNumBits(NumBits numBits);
  inline const NumBits getNumBits();

  inline void setBytesPerBlock(int blocksize);
  inline const int getBytesPerBlock();

  int operator&(AudioFormat &right);

  //Utility functions for converting integer values
  //to enumerated types.
  static NumBits intToNumBits(int numBits);
  static Channel intToChannel(int channel);
  static SampleRate intToSampleRate(int sampleRate);

  protected:

  SampleRate sampleRate;
  Encoding encoding;
  Channel channel;
  NumBits numBits;
  int bytesPerBlock;

  private:
};

void 
AudioFormat::setNumBits(NumBits numBits)
{
  this->numBits = numBits;
}

const AudioFormat::NumBits
AudioFormat::getNumBits()
{
  return numBits;
}

const AudioFormat::Encoding 
AudioFormat::getEncoding()
{
  return encoding;
}

void 
AudioFormat::setEncoding(AudioFormat::Encoding encoding)
{
  this->encoding = encoding;
}


const AudioFormat::SampleRate 
AudioFormat::getSampleRate()
{
  return sampleRate;
}

void 
AudioFormat::setSampleRate(AudioFormat::SampleRate samplerate)
{
  this->sampleRate = samplerate;
}

const AudioFormat::Channel 
AudioFormat::getChannel()
{
  return channel;
}

void 
AudioFormat::setChannel(Channel channel)
{
  this->channel = channel;
}

const int 
AudioFormat::getBytesPerBlock()
{
  return bytesPerBlock;
}

void 
AudioFormat::setBytesPerBlock(int blocksize)
{
  this->bytesPerBlock = blocksize;
}
#endif