// RTP.cpp: implementation of the RTP class.
//
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#ifdef WIN32
#pragma warning (disable:4786)
#endif

#include "RTP.h"
#include "RTPException.h"
#include "RTPEvent.h"
#include "RTPSession.h"

#include <Utils/Guard.h>

#include <time.h>

// Global schedule function required by RTP library.  This calls
// RTP::schedule.
void RTPSchedule(context cid, rtp_opaque_t opaque, timeval *tp);

//initialize single rtp instance
RTP* RTP::pRTP = NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// RTP::RTP()
RTP::RTP()
{
  //Nothing to do.
}//RTP::RTP

// RTP::~RTP()
// Destructor.  Destroys any existing sessions.
RTP::~RTP()
{
  map<context,RTPSession *>::iterator itr;
  
  RTPSession* session = NULL;
  
  // empty the session list.
  for(itr = sessions.begin();itr != sessions.end();itr++){
    session = (*itr).second;
    
    //stops the session's thread.
    session->stop();
    
    delete session;
  }
  
  sessions.clear();
  
}//RTP::~RTP

// RTP::instance()
// Retrieves the single global instance.  The instance
// is created if it does not already exist.
RTP* RTP::instance()
{
  //if not already created
  if(pRTP == NULL){
    
    //get creation mutex
    createMutex.acquire();
    
    //make sure we were first
    if(pRTP == NULL){
      //create the RTP instance and start its thread
      pRTP = new RTP();
    }
    
    //release creation mutex
    createMutex.release();
  }
  
  return pRTP;
}//RTP::instance

// RTP::create(string &host, int port, int ttl)
// Creates a new RTP session instance using the given parameters.
// The new session is then added to the sesion list, and its thread
// is started.
RTPSession* RTP::create(string &host, int port, int ttl)
{
  
  //create the new session
  RTPSession *session = new RTPSession(host,port,ttl);
  
  Guard<Mutex> sessionGuard(sessionMutex);
  
  //add it to the session list
  sessions[session->getContext()] = session;
  
  //kick off the session thread
  session->start();
  
  return session;
}//RTP::create

// RTP::destroy(RTPSession* session)
// Deallocates and removes the given instance from the session list.
void RTP::destroy(RTPSession* session)
{
  Guard<Mutex> sessionGuard(sessionMutex);
  
  sessions.erase(session->getContext());
  
  delete session;
}//RTP::destroy

// RTP::schedule(context cid, rtp_opaque_t opaque, timeval *tp)
// Calls the schedule function of the appropriate RTPSession instance
void RTP::schedule(context cid, rtp_opaque_t opaque, timeval *tp)
{
  
  //wrap the rtp event in an object
  RTPEvent event(cid,opaque,tp);
  
  //acquire mutex lock
  Guard<Mutex> mutexGuard(sessionMutex);
  
  //tell session of new event
  sessions[cid]->schedule(event);
}//RTP::schedule

//global callback function used by the RTP library to schedule events
void RTPSchedule(context cid, rtp_opaque_t opaque, timeval *tp)
{
  //call static the class member to do the real work
  RTP::instance()->schedule(cid,opaque,tp);
}//RTPSchedule