// RTP.h: interface for the RTP class.  Not instantiated directly.  
// This class follows the singleton pattern.  Use RTP to create and
// destroy rtp sessions.  
// The global RTPSchedule function is defined
// in RTP.cpp.  This simply calls RTP::schedule function.
//
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#if !defined(_RTP_H_)
#define _RTP_H_

#include <string>
#include <queue>
#include <map>

using namespace std;

#include "RTPCommon.h"

#include "RTPEvent.h"

#include <Utils/Mutex.h>

class RTPSession;

class RTP
{
public:
  
  // Function that is called by RTP library to schedule new events.
  // This function is called by the global RTPSchedule function defined
  // in RTP.cpp.
  // See RTP documentation for description of parameters.
  void schedule(context cid, rtp_opaque_t opaque, struct timeval *tp);
  
  // Destroys an existing RTPSession, throws RTPException on error
  // @param session The session to destroy.
  // @throw RTPException on error
  static void destroy(RTPSession* session);
  
  // Creates a new RTPSession, throws RTPException on error.
  // @param host Address of the remote host.  This may be either  
  //    dotted quad (12.216.43.21) or hostname (rtp.cu.edu).
  // @param port The destination port number, must be nonzero
  // @param ttl The TTL value for multicast packets.  Ignored in unicast.
  // @return session New session pointer is returned through
  // this parameter.
  // @throw RTPException on error
  static RTPSession* create(string &host, int port, int ttl);
  
  // Returns a pointer to the single RTP instance.
  // @return single RTP instance.
  static RTP* instance();
  
  
  virtual ~RTP();
  
protected:
  // Constructor.
  RTP();
  
  //the current system time in seconds
  double currentTime;
  
  // Table of the active RTP Sessions.  Sessions are added by calling
  // create, and removed by calling destroy.
  static map<context,RTPSession *> sessions;
  
  //single RTP instance
  static RTP* pRTP;
  
  //mutex to make sure only one instance is created.
  static Mutex createMutex;
  
  //mutex to protect static session map
  static Mutex sessionMutex;
};

#endif // !defined(_RTP_H_)
