// RTPSession.cpp: implementation of the RTPSession class.
//
// Many of these functions are an attempt to adapt the RTP library
// to an Object-oriented, multi-threaded environment.  As such they
// are thin wrappers around the actual RTP library calls.  They do
// not include the context id as a parameter and the are protected by a
// mutex.
//
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#ifdef WIN32
#pragma warning (disable:4786)
#endif

#include <Utils/Guard.h>
#include <stdio.h>

#include "RTPSession.h"
#include "RTPException.h"
#include "RTPEvent.h"

// RTPSession::RTPSession(string &host, int port, int ttl)
// Creates a new rtp session by making the appropriate RTP library calls.
RTPSession::RTPSession(string &host, int port, int ttl)
:host(host),port(port),ttl(ttl)
{
  int rc;	//holder for return code
  
  rc = ::RTPCreate(&cid);
  
  if(rc != RTP_OK)
    throw RTPException(rc,
      string("RTPSession::RTPSession: error on call to RTPCreate")
    );
  
  rc = ::RTPSessionAddSendAddr(cid, (char*)host.c_str(), port, ttl);
  
  if(rc != RTP_OK)
    throw RTPException(rc,
      string("RTPSession::RTPSession: error on call to RTPSessionAddSendAddr")
    );
  
  //create canonical name
  string cname("user");
  char hstname[256];
  gethostname(hstname,256);
  
  cname.append("@");
  cname.append(hstname);
  
  rc = ::RTPMemberInfoSetSDES(cid, 0, RTP_MI_CNAME, (char*)cname.c_str());
  
  if(rc != RTP_OK)
    throw RTPException(rc,
      string("RTPSession::RTPSession: error on call to RTPMemberInfoSetSDES"));
  
  rc = ::RTPMemberInfoSetSDES(cid, 0, RTP_MI_NAME, "RTP Example Server");
  
  if(rc != RTP_OK)
    throw RTPException(rc,
      string("RTPSession::RTPSession: error on call to RTPMemberInfoSetSDES"));
  
  rc = ::RTPOpenConnection(cid);
  
  if(rc != RTP_OK)
    throw RTPException(rc,
      string("RTPSession::RTPSession: error on call to RTPOpenConnection"));
}//RTPSession::RTPSession

// RTPSession::~RTPSession()
// Destroys this rtp session by making the appropriate rtp library calls.
RTPSession::~RTPSession()
{
  int rc;	//holder for return code
  rc = ::RTPCloseConnection(cid,"Goodbye!");
  if(rc != RTP_OK)
    throw RTPException(rc,
      string("RTPSession::~RTPSession: error on call to RTPCloseConnection"));
  
  rc = ::RTPDestroy(cid);
  if(rc != RTP_OK)
    throw RTPException(rc,
      string("RTPSession::~RTPSession: error on call to RTPDestroy"));
}//RTPSession::~RTPSession

// RTPSession::send(int tsinc, char marker, short pti, char *payload, int len)
// Sends data for this rtp session.
int RTPSession::send(int tsinc, char marker, short pti, char *payload, int len)
{
  int rc; //holder for return code
  
  Guard<Mutex> sessionGuard(sessionMutex);
  
  rc = ::RTPSend(cid,tsinc,marker,pti,(signed char*)payload,len);
  
  if(rc != RTP_OK)
    throw RTPException(rc,"RTPSession::send: error on call to RTPSend");
  
  return RTP_OK;
}//RTPSession::send

// RTPSession::getRTPSocket(socktype *socket)
// returns the rtp socket descriptor.
int RTPSession::getRTPSocket(socktype *socket)
{
  int rc; //holder for return code
  
  Guard<Mutex> sessionGuard(sessionMutex);
  
  rc = ::RTPSessionGetRTPSocket(cid,socket);
  
  if(rc != RTP_OK)
    throw RTPException(rc,
      "RTPSession::getRTPSocket: error on call to RTPSessionGetRTPSocket");
  
  return RTP_OK;
}//RTPSession::getRTPSocket

// RTPSession::setBandwidth(float bandwidth, float rtcpFraction)
// Sets this sessions bandwidth by calling RTPSessionSetBandwidth
int RTPSession::setBandwidth(float bandwidth, float rtcpFraction)
{
  int rc; //holder for return code
  
  Guard<Mutex> sessionGuard(sessionMutex);
  
  rc = ::RTPSessionSetBandwidth(cid, bandwidth, rtcpFraction);
  
  if(rc != RTP_OK)
    throw RTPException(rc,
      "RTPSession::setBandwidth: error on call to RTPSessionSetBandwidth");
  
  return RTP_OK;
}//RTPSession::setBandwidth

// RTPSession::execute(RTPEvent &event)
// Calls the rtp library call, RTPExecute.
int RTPSession::execute(RTPEvent &event)
{
  
  Guard<Mutex> sessionGuard(sessionMutex);
  
  int rc = ::RTPExecute(event.getContextID(),event.getOpaque());
  
  return rc;
}//RTPSession::execute

// RTPSession::getRTCPSocket(socktype *socket)
// returns the rtcp socket descriptor.
int RTPSession::getRTCPSocket(socktype *socket)
{
  int rc; //holder for return code
  
  Guard<Mutex> sessionGuard(sessionMutex);
  
  rc = ::RTPSessionGetRTCPSocket(cid,socket);
  
  if(rc != RTP_OK)
    throw RTPException(rc,
      "RTPSession::getRTCPSocket: error on call to RTPSessionGetRTCPSocket");
  
  return RTP_OK;
}//RTPSession::getRTCPSocket

// RTPSession::exec()
// The thread for this session.  Based on the main function in the RTP library
// examples.  This thread reads from the rtp, rtcp sockets and executes events
// when they are ready.
void RTPSession::exec()
{
  
  const int NUM_SOCKETS = 2;  //socket for rtp, rtcp
  const int RECEIVE_BUFFER_SIZE = 65536;  //max size of udp packet
  
  timeval timeout_tv; //timeout for select
  double timeout=0.0;
  timeval now_tv; //current time
  double now;
  
  RTPEvent event;
  
  char recvbuf[RECEIVE_BUFFER_SIZE];
  
  fd_set sockets;
  socktype sock[NUM_SOCKETS];
  
  getRTPSocket(&(sock[0]));
  getRTCPSocket(&(sock[1]));
  
  FD_ZERO(&sockets);
  FD_SET(sock[0], &sockets);
  FD_SET(sock[1], &sockets);
  
  int nfds=0;
  
  bool eventscheduled = false; //indicates if an event has been scheduled
  
  while(true){
    
    //check for stop signal
    handleStop();
    
    //get current time
    gettimeofday(&now_tv, NULL);
    now = tv2dbl(now_tv);
    
    queueMutex.acquire();
    
    //check size of queue
    if(eventQueue.size() > 0){
      
      event = eventQueue.front();
      
      eventscheduled = true;
      
    }
    else{
      timeout_tv = dbl2tv(1.0); //wake up at least every second
      eventscheduled = false;
    }
    
    queueMutex.release();
    
    int retval = select(nfds + 1, &sockets, NULL, NULL, &timeout_tv);
    if (retval < 0) { /* select returned an error */
      //perror("select");
      exit(1);
    }
    else if (retval > 0) { /* select reported readable fd's. */
      for (int i = 0; i < 2; i++) {
        if (FD_ISSET(sock[i], &sockets)) {
          int recvbuflen = RECEIVE_BUFFER_SIZE;
          
          Guard<Mutex> sessionGuard(sessionMutex);
          
          int rc = ::RTPReceive(cid, sock[i], recvbuf, &recvbuflen);
          if (rc!= RTP_OK && rc != RTP_PACKET_LOOPBACK) {
            fprintf(stderr, "RTPReceive %s (%d): %s\n",
              i ? "RTCP" : "RTP", sock[i],
              RTPStrError((rtperror)rc));
          }
        }// if (FD_ISSET)
      }// for
    }//else if (retval > 0)
    else { /* retval == 0, select timed out */
      if (eventscheduled) {
        
        gettimeofday(&now_tv, NULL);
        now = tv2dbl(now_tv);
        
        Guard<Mutex> mutexGuard(queueMutex);
        
        //handle all currently ready events
        event = eventQueue.front();
        
        while (eventQueue.empty() != true && event.getTime() <= now) {
          // There is a pending RTP event so run it.
          execute(event);
          
          // pop event off of queue, and get next event
          eventQueue.pop();
          event = eventQueue.front();
        }
        
      }
      else
        continue; /* We timed out, and no events were scheduled so just loop */
    }
  }//while
  
}//RTPSession::exec()

// RTPSession::schedule(RTPEvent &event)
// Adds new event to the event queue.
void RTPSession::schedule(RTPEvent &event)
{
  Guard<Mutex> mutexGuard(queueMutex);
  
  //push it on the event queue for processing
  eventQueue.push(event);
}//RTPSession::schedule
