// RTPSession.h: interface for the RTPSession class.
// Encapsulates functions used for RTP sessions.
// Contains a thread to process events associated with this session.
// 
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#if !defined(_RTPSESSION_H_)
#define _RTPSESSION_H_

#include <string>
#include <queue>

#include "RTPCommon.h"

#include <Utils/Thread.h>
#include <Utils/Mutex.h>

class RTPEvent;

using namespace std;

class RTPSession : public Thread
{
public:
  
  // Schedules an RTP event for processing with this sessions
  // thread.  This function is called from RTP::schedule()
  // @param event An RTPEvent instance that needs processing.
  void schedule(RTPEvent &event);
  
  // Accessor for the RTCP socket
  // @param socket Pointer for the socket value to be returned through.
  // @return RTP_OK
  // @throw RTPException on error.
  int getRTCPSocket(socktype *socket);
  
  // Accessor for the RTP socket
  // @param socket Pointer for the socket value to be returned through.
  // @return RTP_OK
  // @throw RTPException on error.
  int getRTPSocket(socktype *socket);
  
  // Executes an rtp event by calling RTPExecute
  // @param event The event to call RTPExecute for.
  // @return RTP_OK
  // @throw RTPException on error.
  int execute(RTPEvent &event);
  
  // Wrapper for the rtp library RTPSessionSetBandwidth function
  // See the RTP library documentation for details.
  // @return RTP_OK
  // @throw RTPException on error.
  int setBandwidth(float bandwidth, float rtcpFraction);
  
  int send(int tsinc, char marker, short pti, char *payload, int len);
  
  // Create a new RTPSession object.  This constructor creates a new
  // rtp session and establishes a connection with a remote host.
  // @param host Address of the remote host.  This may be either  
  //    dotted quad (12.216.43.21) or hostname (rtp.cu.edu).
  // @param port The destination port number, must be nonzero
  // @param ttl The TTL value for multicast packets.  Ignored in unicast.
  // @throw RTPException if an error occurs.
  RTPSession(string &host, int port, int ttl);
  
  // Destructor.  Calls RTPCloseConnection and RTPDestroy for this session.
  virtual ~RTPSession();
  
  // Accessor for this sessions context id.
  // @return This sessions context id.
  int getContext(){return cid;}
  
protected:
  // Overridden function for thread class
  // This function is responsible for receiveing data
  // from this RTP Session's socket.
  virtual void exec();
  
  //Context id for this session
  context cid;
  
  // The host this session is connected to.
  string host;
  
  // The port this session is connected to.
  int port;
  
  // The time to live value associates with this session.
  int ttl;
  
  // The currently scheduled RTP Events.  Events are scheduled by the RTP
  // library calling the schedule function.  Events are processed in the 
  // _processEvents thread function.  The event queue is protected by
  // a mutex
  queue<RTPEvent> eventQueue;
  
  //protects event queue.
  Mutex queueMutex;
  
  //mutex for access to session data, protects class variables
  //as well as RTP calls.
  Mutex sessionMutex;
};

#endif // !defined(_RTPSESSION_H_)
