// Condition.h: interface for the Condition class.
// Provides a mechanism for one or more threads to block until signaled
// by another thread.  A single thread may be signaled or all waiting threads.
// A condition variable must be protected by a mutex variable.  A thread must
// be activly waiting to receive a signal.  Destroying a condition variable
// before signalling all threads will cause the waiting threads to block 
// forever.
//
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#if !defined(_CONDITION_H_)
#define _CONDITION_H_

#include <list>

using namespace std;

class Semaphore;
class Mutex;

typedef list<Semaphore*> SemaphoreList;

class Condition  
{
public:
  //Signal all threads waiting on this Condition variable.
  //Must have acquired the mutex associated with this variable.
  void signalAll();
  
  //Wait on this variable until it is signaled or the timeout occurs.
  //@param timeout Wait for timeout number of milliseconds.
  //Must have acquired the mutex associated with this variable.
  //for signal.
  //@return bool true if signaled, false otherwise.
  bool wait(int timeout);
  
  //Wait on the variable until it is signaled.
  //Must have acquired the mutex associated with this variable.
  void wait(void);
  
  //Signal one thread waiting on this variable
  //Must have acquired the mutex associated with this variable.
  void signal(void);
  
  //Constructor.  Create a new condition variable.
  //@param mutex The mutex that this variable uses to protect
  //its critical sections.
  Condition(Mutex &mutex);
  
  //Destructor.
  virtual ~Condition();
  
protected:
  //The mutex associated with this variable
  Mutex& mutex;
  
  //The list used to track the threads waiting for
  //this condition variable.
  SemaphoreList semaphoreList;
  
};

#endif // !defined(_CONDITION_H_)
