// Guard.h: defines Guard
//
// Exception safe synchronization class.
// Acquires a resource upon construction, and
// relases it when the instance is destroyed.
// T must define two functions :acquire and relase.
// These functions take no parameters.
//
// Example:
//
// foo(){
//
//  try{
//    //myMutex is acquired
//    Guard<Mutex> mutexGuard(myMutex);
//
//    critical section.
//
//    end of scope for mutexGuard, myMutex is released    
//
//  }
//  catch(exception e){
//
//  }
//
// }//end foo
//
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#if !defined(_GUARD_H_)
#define _GUARD_H_

template <class T> class Guard  
{
public:
  // Creates a new guard object.  Immediately
  // acquires the specified resource.
  // @param mutex Resource to acquire.
  Guard(T& mutex);
  
  // Releases the resource.
  virtual ~Guard();
protected:
  //The resource the guard uses.
  T& mutex;
};

// Guard<T>::Guard(T& mutex)
// acquire the passed mutex.
template <class T> Guard<T>::Guard(T& mutex)
:mutex(mutex)
{
  this->mutex.acquire();
}//Guard<T>::Guard(T& mutex)

// Guard<T>::~Guard()
// release the mutex.
template <class T> Guard<T>::~Guard()
{
  mutex.release();
}//Guard<T>::~Guard()

#endif // !defined(_GUARD_H_)
