// Mutex.cpp: implementation of the Mutex class.
//
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#include "Mutex.h"

#ifdef WIN32
#include <windows.h>
#elif POSIX
#include <pthread.h>
#else
#error port me!
#endif

// Mutex::Mutex()
// Create the os specific mutex handle
Mutex::Mutex()
{
#ifdef WIN32
  hMutex = ::CreateMutex(NULL,false,NULL);
#endif

#ifdef POSIX
  // second arg is null, use default mutex attributes
  ::pthread_mutex_init(&hMutex, NULL);
#endif
  
}//Mutex::Mutex()

// Mutex::~Mutex()
// Close the os specific handle
Mutex::~Mutex()
{
#ifdef WIN32
  ::CloseHandle(hMutex);
#endif

#ifdef POSIX
  ::pthread_mutex_destroy(&hMutex);
#endif
}//Mutex::Mutex()

// Mutex::acquire(int timeout)
// Wait on mutex handle using os specific call.
// Since timeout is not supported with POSIX threads.
// This function simply does a single tryacquire and then
// returns the result.
// Return false if timeout has occured.
bool Mutex::acquire(int timeout)
{

#ifdef WIN32
  int ret = ::WaitForSingleObject(hMutex,timeout);
  
  switch(ret){
  case WAIT_OBJECT_0:
    return true;
    break;
  case WAIT_TIMEOUT:
  default:
    return false;
    break;
  }
  
  return false;
#endif

#ifdef POSIX
  if(::pthread_mutex_trylock(&hMutex) == 1){
    return true;
  }
  else{
    return false;
  }
#endif

}//Mutex::acquire(int timeout)

// Mutex::acquire()
// Wait on mutex handle using os specific call.
// Wait forever.
void Mutex::acquire()
{
#ifdef WIN32
  ::WaitForSingleObject(hMutex,INFINITE);
#endif

#ifdef POSIX
  ::pthread_mutex_lock( &hMutex );
#endif
}//Mutex::acquire()

// Mutex::release()
// Relase mutex handle using os specific call
void Mutex::release()
{
#ifdef WIN32
  ::ReleaseMutex(hMutex);
#endif

#ifdef POSIX
  ::pthread_mutex_unlock( &hMutex );
#endif
}//Mutex::release()


