// Semaphore.cpp: implementation of the Semaphore class.
//
// Posix does not provide a semaphore object.  This implementation
// uses a mutex and a condition variable for posix platforms.
//
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#include "Semaphore.h"

// Semaphore::Semaphore(int count)
// Create os specific semaphore handle.
// Initialize its count to the passed value.
Semaphore::Semaphore(int count)
#ifdef POSIX
:count(count)
#endif
{
#ifdef WIN32
  hSemaphore = ::CreateSemaphore(NULL, count, 10000, NULL);
#endif

#ifdef POSIX
  ::pthread_mutex_init(&hMutex, NULL);
  ::pthread_cond_init(&hCondition, NULL);
#endif
}//Semaphore::Semaphore(int count)

// Semaphore::~Semaphore()
// Close the os semaphore handle
Semaphore::~Semaphore()
{
#ifdef WIN32
  ::CloseHandle(hSemaphore);
#endif

#ifdef POSIX
  ::pthread_mutex_destroy(&hMutex);
  ::pthread_cond_destroy(&hCondition);
#endif
}//Semaphore::~Semaphore()

// Semaphore::acquire()
// Wait forever to acquire the semaphore
void Semaphore::acquire()
{
#ifdef WIN32
  ::WaitForSingleObject(hSemaphore,INFINITE);
#endif

#ifdef POSIX
  ::pthread_mutex_lock(&hMutex);

  while (count <= 0)
  {
   ::pthread_cond_wait (&hCondition, &hMutex);
  }

  count--;

  ::pthread_mutex_unlock(&hMutex);
#endif
}//Semaphore::acquire()

// Semaphore::release()
// Relase the os semaphore handle
void Semaphore::release()
{
#ifdef WIN32
  ::ReleaseSemaphore(hSemaphore,1,NULL);
#endif

#ifdef POSIX
  ::pthread_mutex_lock(&hMutex);

  count++;

  ::pthread_mutex_unlock(&hMutex);

  ::pthread_cond_signal(&hCondition);
#endif
}//Semaphore::release()

// Semaphore::acquire(int timeout)
// Wait until timeout to acquire the semaphore.
// Return false if timeout occurs.
// since timeout is unsupported on posix systems
// this function simply checks the count and only
// returns true if the count is > 0
bool Semaphore::acquire(int timeout)
{
#ifdef WIN32
  int ret = WaitForSingleObject(hSemaphore,timeout);
  
  //see if we acquired the semaphore
  switch(ret){
  case WAIT_OBJECT_0:
    return true;
    break;
  case WAIT_TIMEOUT:
  default:
    return false;
    break;
  }
  
  return false;
#endif

#ifdef POSIX
  bool ret = false;

  ::pthread_mutex_lock(&hMutex);

  if(count > 0)
  {
    count--;
    ret = true;
  }

  ::pthread_mutex_unlock(&hMutex);

  return ret;
#endif

}//Semaphore::acquire(int timeout)
