// Thread.h: defines Thread, ThreadException, ThreadStoppedException
// Provides a platform independent thread abstraction.  Similar to java's
// thread object.  Derive from this class and override the exec member function.
// Call start to start the thread and stop to stop the thread.
// Call handleStop from within the thread to check for the thread being signaled
// to stop.  If the thread has been signaled, this will throw a ThreadException.
// The exception may be caught and dealtwith.  It must then be rethrown.
//  
//  Example:

//  #include <iostream>
//  #include <Thread.h>
//
//  using namespace std;
//  
//  class X : public Thread{
//  
//  protected:
//    void exec(){
//      while(true){
//        try{
//          handleStop();
//        }
//        catch(ThreadException &e){
//          cout << "Stopping..." << e.what() << endl;
//          break;
//        }
//        
//        cout << "Running..." << endl;
//
//      }
//    }
//  };
//
//  int main()
//  {
//
//    //create thread class
//    X x;
//    //strart the thread
//    x.start();
//  
//    //allow thread to run
//    char y;
//    cin >> y;
//
//    x.stop();
//  
//    return 0;
//  }
//
//
// Copyright (c) 2001 by Columbia University. All rights reserved.
//////////////////////////////////////////////////////////////////////

#if !defined(_THREAD_H_)
#define _THREAD_H_

#ifdef POSIX
#include <pthread.h>
#endif

#include <exception>

using namespace std;

//  Generic Thread Exception
class ThreadException :public exception{};

//  Thrown when the thread has been signaled to stop
class ThreadStoppedException : public ThreadException
{
  const char *what() const throw(){
    return "User requested that thread stop execution.";
  }
};


class Thread  
{
public:
  //  Resumes this thread
  void resume();
  
  //  Pauses this thread
  void pause();
  
  //  Stops this thread
  void stop();
  
  //  starts this thread
  void start();
  
  // Destructor
  virtual ~Thread();

  // Constructor
  Thread();

protected:
  // flag indicating whether or not this thread has been paused
  bool paused;

  // flag indicating whether or not this thread has been stopped
  bool stopped;

  // thread handle
#ifdef POSIX
  pthread_t threadHandle;
#endif
  
  //  Checks the stopped flag.  Throws a ThreadStoppedException
  //  if the stop method has been called.
  void handleStop(); //throw(ThreadStoppedException);
  
  //  !! NOT IMPLEMENTED !!
  //  Checks the paused flag.  Blocks execution until the resume
  //  method is called.
  void handlePause();
  
  //  Virtual function to be overridden by subclass.
  //  This function is called by startThread to do 
  //  the real work.
  virtual void exec(void) = 0;
  
  //  Passed to the OS specific thread creation function.
  //  It calls exec on the passed pointer.
  //  @param pThread Pointer to a thread object.
  static void startThread(void* pThread);
};

#endif // !defined(_THREAD_H_)
